@@ -1,19 +1,16 @@
-;;; magit-apply.el --- apply Git diffs  -*- lexical-binding: t -*-
+;;; magit-apply.el --- Apply Git diffs  -*- lexical-binding:t -*-
 
-;; Copyright (C) 2010-2021  The Magit Project Contributors
-;;
-;; You should have received a copy of the AUTHORS.md file which
-;; lists all contributors.  If not, see http://magit.vc/authors.
+;; Copyright (C) 2008-2024 The Magit Project Contributors
 
-;; Author: Jonas Bernoulli <jonas@bernoul.li>
-;; Maintainer: Jonas Bernoulli <jonas@bernoul.li>
+;; Author: Jonas Bernoulli <emacs.magit@jonas.bernoulli.dev>
+;; Maintainer: Jonas Bernoulli <emacs.magit@jonas.bernoulli.dev>
 
 ;; SPDX-License-Identifier: GPL-3.0-or-later
 
-;; Magit is free software; you can redistribute it and/or modify it
+;; Magit is free software: you can redistribute it and/or modify it
 ;; under the terms of the GNU General Public License as published by
-;; the Free Software Foundation; either version 3, or (at your option)
-;; any later version.
+;; the Free Software Foundation, either version 3 of the License, or
+;; (at your option) any later version.
 ;;
 ;; Magit is distributed in the hope that it will be useful, but WITHOUT
 ;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
@@ -21,7 +18,7 @@
 ;; License for more details.
 ;;
 ;; You should have received a copy of the GNU General Public License
-;; along with Magit.  If not, see http://www.gnu.org/licenses.
+;; along with Magit.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Commentary:
 
@@ -39,8 +36,8 @@
 (require 'transient) ; See #3732.
 
 ;; For `magit-apply'
-(declare-function magit-am "magit-sequence" ())
-(declare-function magit-patch-apply "magit-files" ())
+(declare-function magit-am "magit-sequence" () t)
+(declare-function magit-patch-apply "magit-patch" () t)
 ;; For `magit-discard-files'
 (declare-function magit-checkout-stage "magit-merge" (file arg))
 (declare-function magit-checkout-read-stage "magit-merge" (file))
@@ -50,9 +47,6 @@
                   (url &optional path name args))
 (declare-function magit-submodule-read-name-for-path "magit-submodule"
                   (path &optional prefer-short))
-(declare-function borg--maybe-absorb-gitdir "borg" (pkg))
-(declare-function borg--sort-submodule-sections "borg" (file))
-(declare-function borg-assimilate "borg" (package url &optional partially))
 (defvar borg-user-emacs-directory)
 
 ;;; Options
@@ -104,9 +98,6 @@ is a member of `magit-post-stage-hook-commands'."
   :group 'magit-commands
   :type 'hook)
 
-(defvar magit-post-stage-hook-commands
-  '(magit-stage magit-stage-file magit-stage-modified))
-
 (defcustom magit-post-unstage-hook nil
   "Hook run after unstaging changes.
 This hook is run by `magit-refresh' if `this-command'
@@ -115,9 +106,6 @@ is a member of `magit-post-unstage-hook-commands'."
   :group 'magit-commands
   :type 'hook)
 
-(defvar magit-post-unstage-hook-commands
-  '(magit-unstage magit-unstage-file magit-unstage-all))
-
 ;;; Commands
 ;;;; Apply
 
@@ -126,19 +114,19 @@ is a member of `magit-post-unstage-hook-commands'."
 With a prefix argument fallback to a 3-way merge.  Doing
 so causes the change to be applied to the index as well."
   (interactive (and current-prefix-arg (list "--3way")))
-  (--when-let (magit-apply--get-selection)
+  (when-let ((s (magit-apply--get-selection)))
     (pcase (list (magit-diff-type) (magit-diff-scope))
-      (`(,(or `unstaged `staged) ,_)
+      (`(,(or 'unstaged 'staged) ,_)
        (user-error "Change is already in the working tree"))
-      (`(untracked ,(or `file `files))
-       (call-interactively 'magit-am))
-      (`(,_ region) (magit-apply-region it args))
-      (`(,_   hunk) (magit-apply-hunk   it args))
-      (`(,_  hunks) (magit-apply-hunks  it args))
+      (`(untracked ,(or 'file 'files))
+       (call-interactively #'magit-am))
+      (`(,_ region) (magit-apply-region s args))
+      (`(,_   hunk) (magit-apply-hunk   s args))
+      (`(,_  hunks) (magit-apply-hunks  s args))
       (`(rebase-sequence file)
-       (call-interactively 'magit-patch-apply))
-      (`(,_   file) (magit-apply-diff   it args))
-      (`(,_  files) (magit-apply-diffs  it args)))))
+       (call-interactively #'magit-patch-apply))
+      (`(,_   file) (magit-apply-diff   s args))
+      (`(,_  files) (magit-apply-diffs  s args)))))
 
 (defun magit-apply--section-content (section)
   (buffer-substring-no-properties (if (magit-hunk-section-p section)
@@ -192,38 +180,39 @@ adjusted as \"@@ -10,6 +10,7 @@\" and \"@@ -18,6 +19,7 @@\"."
 (defun magit-apply--adjust-hunk-new-start (hunk)
   (car (magit-apply--adjust-hunk-new-starts (list hunk))))
 
-(defun magit-apply-hunks (sections &rest args)
-  (let ((section (oref (car sections) parent)))
-    (when (string-match "^diff --cc" (oref section value))
+(defun magit-apply-hunks (hunks &rest args)
+  (let ((file (oref (car hunks) parent)))
+    (when (magit-diff--combined-p file)
       (user-error "Cannot un-/stage resolution hunks.  Stage the whole file"))
     (magit-apply-patch
-     section args
-     (concat (oref section header)
-             (mapconcat #'identity
-                        (magit-apply--adjust-hunk-new-starts
-                         (mapcar #'magit-apply--section-content sections))
-                        "")))))
+     file args
+     (concat (oref file header)
+             (string-join (magit-apply--adjust-hunk-new-starts
+                           (mapcar #'magit-apply--section-content hunks)))))))
 
-(defun magit-apply-hunk (section &rest args)
-  (when (string-match "^diff --cc" (magit-section-parent-value section))
-    (user-error "Cannot un-/stage resolution hunks.  Stage the whole file"))
-  (let* ((header (car (oref section value)))
-         (header (and (symbolp header) header))
-         (content (magit-apply--section-content section)))
+(defun magit-apply-hunk (hunk &rest args)
+  (let ((file (oref hunk parent)))
+    (when (magit-diff--combined-p file)
+      (user-error "Cannot un-/stage resolution hunks.  Stage the whole file"))
+    (let* ((header (car (oref hunk value)))
+           (header (and (symbolp header) header))
+           (content (magit-apply--section-content hunk)))
+      (magit-apply-patch
+       file args
+       (concat (magit-diff-file-header hunk (not (eq header 'rename)))
+               (if header
+                   content
+                 (magit-apply--adjust-hunk-new-start content)))))))
+
+(defun magit-apply-region (hunk &rest args)
+  (let ((file (oref hunk parent)))
+    (when (magit-diff--combined-p file)
+      (user-error "Cannot un-/stage resolution hunks.  Stage the whole file"))
     (magit-apply-patch
-     (oref section parent) args
-     (concat (magit-diff-file-header section (not (eq header 'rename)))
-             (if header
-                 content
-               (magit-apply--adjust-hunk-new-start content))))))
-
-(defun magit-apply-region (section &rest args)
-  (when (string-match "^diff --cc" (magit-section-parent-value section))
-    (user-error "Cannot un-/stage resolution hunks.  Stage the whole file"))
-  (magit-apply-patch (oref section parent) args
-                     (concat (magit-diff-file-header section)
-                             (magit-apply--adjust-hunk-new-start
-                              (magit-diff-hunk-region-patch section args)))))
+     file args
+     (concat (magit-diff-file-header hunk)
+             (magit-apply--adjust-hunk-new-start
+              (magit-diff-hunk-region-patch hunk args))))))
 
 (defun magit-apply-patch (section:s args patch)
   (let* ((files (if (atom section:s)
@@ -253,9 +242,9 @@ adjusted as \"@@ -10,6 +10,7 @@\" and \"@@ -18,6 +19,7 @@\"."
   (or (magit-region-sections '(hunk file module) t)
       (let ((section (magit-current-section)))
         (pcase (oref section type)
-          ((or `hunk `file `module) section)
-          ((or `staged `unstaged `untracked
-               `stashed-index `stashed-worktree `stashed-untracked)
+          ((or 'hunk 'file 'module) section)
+          ((or 'staged 'unstaged 'untracked
+               'stashed-index 'stashed-worktree 'stashed-untracked)
            (oref section children))
           (_ (user-error "Cannot apply this, it's not a change"))))))
 
@@ -269,14 +258,23 @@ adjusted as \"@@ -10,6 +10,7 @@\" and \"@@ -18,6 +19,7 @@\"."
             sections))
     (t sections)))
 
-(defun magit-apply--diff-ignores-whitespace-p ()
-  (and (cl-intersection magit-buffer-diff-args
-                        '("--ignore-space-at-eol"
-                          "--ignore-space-change"
-                          "--ignore-all-space"
-                          "--ignore-blank-lines")
-                        :test #'equal)
-       t))
+(defun magit-apply--ignore-whitespace-p (selection type scope)
+  "Return t if it is necessary and possible to ignore whitespace.
+It is necessary to do so when the diff ignores whitespace changes
+and whole files are being applied.  It is possible when no binary
+files are involved.  If it is both necessary and impossible, then
+return nil, possibly causing whitespace changes to be applied."
+  (and (memq type  '(unstaged staged))
+       (memq scope '(file files list))
+       (cl-find-if (lambda (arg)
+                     (member arg '("--ignore-space-at-eol"
+                                   "--ignore-space-change"
+                                   "--ignore-all-space"
+                                   "--ignore-blank-lines")))
+                   magit-buffer-diff-args)
+       (not (cl-find-if (lambda (section)
+                          (oref section binary))
+                        (ensure-list selection)))))
 
 ;;;; Stage
 
@@ -285,43 +283,61 @@ adjusted as \"@@ -10,6 +10,7 @@\" and \"@@ -18,6 +19,7 @@\"."
 With a prefix argument, INTENT, and an untracked file (or files)
 at point, stage the file but not its content."
   (interactive "P")
-  (--if-let (and (derived-mode-p 'magit-mode) (magit-apply--get-selection))
-      (pcase (list (magit-diff-type)
-                   (magit-diff-scope)
-                   (magit-apply--diff-ignores-whitespace-p))
+  (if-let ((s (and (derived-mode-p 'magit-mode)
+                   (magit-apply--get-selection)))
+           (type (magit-diff-type))
+           (scope (magit-diff-scope)))
+      (pcase (list type scope
+                   (magit-apply--ignore-whitespace-p s type scope))
         (`(untracked     ,_  ,_) (magit-stage-untracked intent))
-        (`(unstaged  region  ,_) (magit-apply-region it "--cached"))
-        (`(unstaged    hunk  ,_) (magit-apply-hunk   it "--cached"))
-        (`(unstaged   hunks  ,_) (magit-apply-hunks  it "--cached"))
-        (`(unstaged    file   t) (magit-apply-diff   it "--cached"))
-        (`(unstaged   files   t) (magit-apply-diffs  it "--cached"))
-        (`(unstaged    list   t) (magit-apply-diffs  it "--cached"))
-        (`(unstaged    file nil) (magit-stage-1 "-u" (list (oref it value))))
-        (`(unstaged   files nil) (magit-stage-1 "-u" (magit-region-values nil t)))
-        (`(unstaged    list nil) (magit-stage-modified))
+        (`(unstaged  region  ,_) (magit-apply-region s "--cached"))
+        (`(unstaged    hunk  ,_) (magit-apply-hunk   s "--cached"))
+        (`(unstaged   hunks  ,_) (magit-apply-hunks  s "--cached"))
+        ('(unstaged    file   t) (magit-apply-diff   s "--cached"))
+        ('(unstaged   files   t) (magit-apply-diffs  s "--cached"))
+        ('(unstaged    list   t) (magit-apply-diffs  s "--cached"))
+        ('(unstaged    file nil) (magit-stage-1 "-u" (list (oref s value))))
+        ('(unstaged   files nil) (magit-stage-1 "-u" (magit-region-values nil t)))
+        ('(unstaged    list nil) (magit-stage-modified))
         (`(staged        ,_  ,_) (user-error "Already staged"))
         (`(committed     ,_  ,_) (user-error "Cannot stage committed changes"))
         (`(undefined     ,_  ,_) (user-error "Cannot stage this change")))
-    (call-interactively 'magit-stage-file)))
+    (call-interactively #'magit-stage-file)))
 
 ;;;###autoload
-(defun magit-stage-file (file)
-  "Stage all changes to FILE.
-With a prefix argument or when there is no file at point ask for
-the file to be staged.  Otherwise stage the file at point without
-requiring confirmation."
-  (interactive
-   (let* ((atpoint (magit-section-value-if 'file))
-          (current (magit-file-relative-name))
-          (choices (nconc (magit-unstaged-files)
-                          (magit-untracked-files)))
-          (default (car (member (or atpoint current) choices))))
-     (list (if (or current-prefix-arg (not default))
-               (magit-completing-read "Stage file" choices
-                                      nil t nil nil default)
-             default))))
+(defun magit-stage-buffer-file ()
+  "Stage all changes to the file being visited in the current buffer."
+  (interactive)
+  (unless buffer-file-name
+    (user-error "Not visiting a file"))
   (magit-with-toplevel
-    (magit-stage-1 nil (list file))))
+    (magit-stage-1 (and (magit-file-ignored-p buffer-file-name)
+                        (if (y-or-n-p "Visited file is ignored; stage anyway?")
+                            "--force"
+                          (user-error "Abort")))
+                   (list (magit-file-relative-name)))))
+
+;;;###autoload
+(defun magit-stage-file (files &optional force)
+  "Read one or more files and stage all changes in those files.
+With prefix argument FORCE, offer ignored files for completion."
+  (interactive
+   (let* ((choices (if current-prefix-arg
+                       (magit-ignored-files)
+                     (nconc (magit-unstaged-files)
+                            (magit-untracked-files))))
+          (default (or (magit-section-value-if 'file)
+                       (magit-file-relative-name)))
+          (default (car (member default choices))))
+     (list (magit-completing-read-multiple
+            (if current-prefix-arg "Stage ignored file,s: " "Stage file,s: ")
+            choices nil t nil nil default)
+           current-prefix-arg)))
+  (magit-with-toplevel
+    ;; For backward compatibility, and because of
+    ;; the function's name, don't require a list.
+    (magit-stage-1 (and force "--force")
+                   (ensure-list files))))
 
 ;;;###autoload
 (defun magit-stage-modified (&optional all)
@@ -346,9 +362,9 @@ ignored) files."
 (defun magit-stage-untracked (&optional intent)
   (let* ((section (magit-current-section))
          (files (pcase (magit-diff-scope)
-                  (`file  (list (oref section value)))
-                  (`files (magit-region-values nil t))
-                  (`list  (magit-untracked-files))))
+                  ('file  (list (oref section value)))
+                  ('files (magit-region-values nil t))
+                  ('list  (magit-untracked-files))))
          plain repos)
     (dolist (file files)
       (if (and (not (file-symlink-p file))
@@ -366,71 +382,90 @@ ignored) files."
         (goto-char (oref (magit-get-section
                           `((file . ,repo) (untracked) (status)))
                          start))
-        (let* ((topdir (magit-toplevel))
-               (url (let ((default-directory
-                            (file-name-as-directory (expand-file-name repo))))
-                      (or (magit-get "remote" (magit-get-some-remote) "url")
-                          (concat (file-name-as-directory ".") repo))))
-               (package
-                (and (equal (bound-and-true-p borg-user-emacs-directory)
-                            topdir)
-                     (file-name-nondirectory (directory-file-name repo)))))
-          (if (and package
-                   (y-or-n-p (format "Also assimilate `%s' drone?" package)))
-              (borg-assimilate package url)
-            (magit-submodule-add-1
-             url repo (magit-submodule-read-name-for-path repo package))
-            (when package
-              (borg--sort-submodule-sections
-               (expand-file-name ".gitmodules" topdir))
-              (let ((default-directory borg-user-emacs-directory))
-                (borg--maybe-absorb-gitdir package)))))))
+        (when (and (fboundp 'borg-assimilate)
+                   (fboundp 'borg--maybe-absorb-gitdir)
+                   (fboundp 'borg--sort-submodule-sections))
+          (let* ((topdir (magit-toplevel))
+                 (url (let ((default-directory
+                             (file-name-as-directory (expand-file-name repo))))
+                        (or (magit-get "remote" (magit-get-some-remote) "url")
+                            (concat (file-name-as-directory ".") repo))))
+                 (package
+                  (and (equal borg-user-emacs-directory topdir)
+                       (file-name-nondirectory (directory-file-name repo)))))
+            (if (and package
+                     (y-or-n-p (format "Also assimilate `%s' drone?" package)))
+                (borg-assimilate package url)
+              (magit-submodule-add-1
+               url repo (magit-submodule-read-name-for-path repo package))
+              (when package
+                (borg--sort-submodule-sections
+                 (expand-file-name ".gitmodules" topdir))
+                (let ((default-directory borg-user-emacs-directory))
+                  (borg--maybe-absorb-gitdir package))))))))
     (magit-wip-commit-after-apply files " after stage")))
 
+(defvar magit-post-stage-hook-commands
+  '(magit-stage
+    magit-stage-buffer-file
+    magit-stage-file
+    magit-stage-modified))
+
+(defun magit-run-post-stage-hook ()
+  (when (memq this-command magit-post-stage-hook-commands)
+    (magit-run-hook-with-benchmark 'magit-post-stage-hook)))
+
 ;;;; Unstage
 
 (defun magit-unstage ()
   "Remove the change at point from the staging area."
   (interactive)
-  (--when-let (magit-apply--get-selection)
-    (pcase (list (magit-diff-type)
-                 (magit-diff-scope)
-                 (magit-apply--diff-ignores-whitespace-p))
+  (when-let ((s (magit-apply--get-selection))
+             (type (magit-diff-type))
+             (scope (magit-diff-scope)))
+    (pcase (list type scope
+                 (magit-apply--ignore-whitespace-p s type scope))
       (`(untracked     ,_  ,_) (user-error "Cannot unstage untracked changes"))
-      (`(unstaged    file  ,_) (magit-unstage-intent (list (oref it value))))
+      (`(unstaged    file  ,_) (magit-unstage-intent (list (oref s value))))
       (`(unstaged   files  ,_) (magit-unstage-intent (magit-region-values nil t)))
       (`(unstaged      ,_  ,_) (user-error "Already unstaged"))
-      (`(staged    region  ,_) (magit-apply-region it "--reverse" "--cached"))
-      (`(staged      hunk  ,_) (magit-apply-hunk   it "--reverse" "--cached"))
-      (`(staged     hunks  ,_) (magit-apply-hunks  it "--reverse" "--cached"))
-      (`(staged      file   t) (magit-apply-diff   it "--reverse" "--cached"))
-      (`(staged     files   t) (magit-apply-diffs  it "--reverse" "--cached"))
-      (`(staged      list   t) (magit-apply-diffs  it "--reverse" "--cached"))
-      (`(staged      file nil) (magit-unstage-1 (list (oref it value))))
-      (`(staged     files nil) (magit-unstage-1 (magit-region-values nil t)))
-      (`(staged      list nil) (magit-unstage-all))
+      (`(staged    region  ,_) (magit-apply-region s "--reverse" "--cached"))
+      (`(staged      hunk  ,_) (magit-apply-hunk   s "--reverse" "--cached"))
+      (`(staged     hunks  ,_) (magit-apply-hunks  s "--reverse" "--cached"))
+      ('(staged      file   t) (magit-apply-diff   s "--reverse" "--cached"))
+      ('(staged     files   t) (magit-apply-diffs  s "--reverse" "--cached"))
+      ('(staged      list   t) (magit-apply-diffs  s "--reverse" "--cached"))
+      ('(staged      file nil) (magit-unstage-1 (list (oref s value))))
+      ('(staged     files nil) (magit-unstage-1 (magit-region-values nil t)))
+      ('(staged      list nil) (magit-unstage-all))
       (`(committed     ,_  ,_) (if magit-unstage-committed
                                    (magit-reverse-in-index)
                                  (user-error "Cannot unstage committed changes")))
       (`(undefined     ,_  ,_) (user-error "Cannot unstage this change")))))
 
 ;;;###autoload
-(defun magit-unstage-file (file)
-  "Unstage all changes to FILE.
-With a prefix argument or when there is no file at point ask for
-the file to be unstaged.  Otherwise unstage the file at point
-without requiring confirmation."
-  (interactive
-   (let* ((atpoint (magit-section-value-if 'file))
-          (current (magit-file-relative-name))
-          (choices (magit-staged-files))
-          (default (car (member (or atpoint current) choices))))
-     (list (if (or current-prefix-arg (not default))
-               (magit-completing-read "Unstage file" choices
-                                      nil t nil nil default)
-             default))))
+(defun magit-unstage-buffer-file ()
+  "Unstage all changes to the file being visited in the current buffer."
+  (interactive)
+  (unless buffer-file-name
+    (user-error "Not visiting a file"))
   (magit-with-toplevel
-    (magit-unstage-1 (list file))))
+    (magit-unstage-1 (list (magit-file-relative-name)))))
+
+;;;###autoload
+(defun magit-unstage-file (files)
+  "Read one or more files and unstage all changes to those files."
+  (interactive
+   (let* ((choices (magit-staged-files))
+          (default (or (magit-section-value-if 'file)
+                       (magit-file-relative-name)))
+          (default (car (member default choices))))
+     (list (magit-completing-read-multiple "Unstage file,s: " choices
+                                           nil t nil nil default))))
+  (magit-with-toplevel
+    ;; For backward compatibility, and because of
+    ;; the function's name, don't require a list.
+    (magit-unstage-1 (ensure-list files))))
 
 (defun magit-unstage-1 (files)
   (magit-wip-commit-before-change files " before unstage")
@@ -458,6 +493,16 @@ without requiring confirmation."
   (magit-run-git "reset" "HEAD" "--" magit-buffer-diff-files)
   (magit-wip-commit-after-apply nil " after unstage"))
 
+(defvar magit-post-unstage-hook-commands
+  '(magit-unstage
+    magit-unstage-buffer-file
+    magit-unstage-file
+    magit-unstage-all))
+
+(defun magit-run-post-unstage-hook ()
+  (when (memq this-command magit-post-unstage-hook-commands)
+    (magit-run-hook-with-benchmark 'magit-post-unstage-hook)))
+
 ;;;; Discard
 
 (defun magit-discard ()
@@ -467,16 +512,16 @@ On a hunk or file with unresolved conflicts prompt which side to
 keep (while discarding the other).  If point is within the text
 of a side, then keep that side without prompting."
   (interactive)
-  (--when-let (magit-apply--get-selection)
+  (when-let ((s (magit-apply--get-selection)))
     (pcase (list (magit-diff-type) (magit-diff-scope))
       (`(committed ,_) (user-error "Cannot discard committed changes"))
       (`(undefined ,_) (user-error "Cannot discard this change"))
-      (`(,_    region) (magit-discard-region it))
-      (`(,_      hunk) (magit-discard-hunk   it))
-      (`(,_     hunks) (magit-discard-hunks  it))
-      (`(,_      file) (magit-discard-file   it))
-      (`(,_     files) (magit-discard-files  it))
-      (`(,_      list) (magit-discard-files  it)))))
+      (`(,_    region) (magit-discard-region s))
+      (`(,_      hunk) (magit-discard-hunk   s))
+      (`(,_     hunks) (magit-discard-hunks  s))
+      (`(,_      file) (magit-discard-file   s))
+      (`(,_     files) (magit-discard-files  s))
+      (`(,_      list) (magit-discard-files  s)))))
 
 (defun magit-discard-region (section)
   (magit-confirm 'discard "Discard region")
@@ -486,8 +531,8 @@ of a side, then keep that side without prompting."
   (magit-confirm 'discard "Discard hunk")
   (let ((file (magit-section-parent-value section)))
     (pcase (cddr (car (magit-file-status file)))
-      (`(?U ?U) (magit-smerge-keep-current))
-      (_ (magit-discard-apply section 'magit-apply-hunk)))))
+      ('(?U ?U) (magit-smerge-keep-current))
+      (_ (magit-discard-apply section #'magit-apply-hunk)))))
 
 (defun magit-discard-apply (section apply)
   (if (eq (magit-diff-type section) 'unstaged)
@@ -503,10 +548,11 @@ of a side, then keep that side without prompting."
       (funcall apply section "--reverse" "--index"))))
 
 (defun magit-discard-hunks (sections)
-  (magit-confirm 'discard (format "Discard %s hunks from %s"
-                                  (length sections)
-                                  (magit-section-parent-value (car sections))))
-  (magit-discard-apply-n sections 'magit-apply-hunks))
+  (magit-confirm 'discard
+    (list "Discard %d hunks from %s"
+          (length sections)
+          (magit-section-parent-value (car sections))))
+  (magit-discard-apply-n sections #'magit-apply-hunks))
 
 (defun magit-discard-apply-n (sections apply)
   (let ((section (car sections)))
@@ -538,16 +584,17 @@ of a side, then keep that side without prompting."
                        (`unstaged ?Y)
                        (`untracked ?Z))
                      (cddr (assoc file status)))
-          (`(?Z) (dolist (f (magit-untracked-files nil file))
+          ('(?Z) (dolist (f (magit-untracked-files nil file))
                    (push f delete)))
-          ((or `(?Z ?? ??) `(?Z ?! ?!)) (push file delete))
-          ((or `(?Z ?D ? ) `(,_ ?D ?D)) (push file delete))
+          ((or '(?Z ?? ??) '(?Z ?! ?!)) (push file delete))
+          ('(?Z ?D ? )                  (push file delete))
+          (`(,_ ?D ?D)                  (push file resolve))
           ((or `(,_ ?U ,_) `(,_ ,_ ?U)) (push file resolve))
           (`(,_ ?A ?A)                  (push file resolve))
           (`(?X ?M ,(or ?  ?M ?D)) (push section discard))
           (`(?Y ,_         ?M    ) (push section discard))
-          (`(?X ?A         ?M    ) (push file discard-new))
-          (`(?X ?C         ?M    ) (push file discard-new))
+          ('(?X ?A         ?M    ) (push file discard-new))
+          ('(?X ?C         ?M    ) (push file discard-new))
           (`(?X ?A ,(or ?     ?D)) (push file delete))
           (`(?X ?C ,(or ?     ?D)) (push file delete))
           (`(?X ?D ,(or ?  ?M   )) (push file resurrect))
@@ -573,7 +620,7 @@ of a side, then keep that side without prompting."
 (defun magit-discard-files--resolve (files)
   (if-let ((arg (and (cdr files)
                      (magit-read-char-case
-                         (format "For these %i files\n%s\ncheckout:\n"
+                         (format "For these %d files\n%s\ncheckout:\n"
                                  (length files)
                                  (mapconcat (lambda (file)
                                               (concat "  " file))
@@ -620,7 +667,7 @@ of a side, then keep that side without prompting."
              (magit-call-git "rm" "--cached" "--force" "--" file))))))
 
 (defun magit-discard-files--rename (files status)
-  (magit-confirm 'rename "Undo rename %s" "Undo %i renames" nil
+  (magit-confirm 'rename "Undo rename %s" "Undo %d renames" nil
     (mapcar (lambda (file)
               (setq file (assoc file status))
               (format "%s -> %s" (cadr file) (car file)))
@@ -629,8 +676,8 @@ of a side, then keep that side without prompting."
     (let ((orig (cadr (assoc file status))))
       (if (file-exists-p file)
           (progn
-            (--when-let (file-name-directory orig)
-              (make-directory it t))
+            (when-let ((path (file-name-directory orig)))
+              (make-directory path t))
             (magit-call-git "mv" file orig))
         (magit-call-git "rm" "--cached" "--" file)
         (magit-call-git "reset" "--" orig)))))
@@ -649,10 +696,10 @@ of a side, then keep that side without prompting."
           (setq sections
                 (--remove (member (oref it value) binaries)
                           sections)))
-        (cond ((= (length sections) 1)
+        (cond ((length= sections 1)
                (magit-discard-apply (car sections) 'magit-apply-diff))
               (sections
-               (magit-discard-apply-n sections 'magit-apply-diffs)))
+               (magit-discard-apply-n sections #'magit-apply-diffs)))
         (when binaries
           (let ((modified (magit-unstaged-files t)))
             (setq binaries (--separate (member it modified) binaries)))
@@ -671,31 +718,31 @@ of a side, then keep that side without prompting."
 With a prefix argument fallback to a 3-way merge.  Doing
 so causes the change to be applied to the index as well."
   (interactive (and current-prefix-arg (list "--3way")))
-  (--when-let (magit-apply--get-selection)
+  (when-let ((s (magit-apply--get-selection)))
     (pcase (list (magit-diff-type) (magit-diff-scope))
       (`(untracked ,_) (user-error "Cannot reverse untracked changes"))
       (`(unstaged  ,_) (user-error "Cannot reverse unstaged changes"))
-      (`(,_    region) (magit-reverse-region it args))
-      (`(,_      hunk) (magit-reverse-hunk   it args))
-      (`(,_     hunks) (magit-reverse-hunks  it args))
-      (`(,_      file) (magit-reverse-file   it args))
-      (`(,_     files) (magit-reverse-files  it args))
-      (`(,_      list) (magit-reverse-files  it args)))))
+      (`(,_    region) (magit-reverse-region s args))
+      (`(,_      hunk) (magit-reverse-hunk   s args))
+      (`(,_     hunks) (magit-reverse-hunks  s args))
+      (`(,_      file) (magit-reverse-file   s args))
+      (`(,_     files) (magit-reverse-files  s args))
+      (`(,_      list) (magit-reverse-files  s args)))))
 
 (defun magit-reverse-region (section args)
   (magit-confirm 'reverse "Reverse region")
-  (magit-reverse-apply section 'magit-apply-region args))
+  (magit-reverse-apply section #'magit-apply-region args))
 
 (defun magit-reverse-hunk (section args)
   (magit-confirm 'reverse "Reverse hunk")
-  (magit-reverse-apply section 'magit-apply-hunk args))
+  (magit-reverse-apply section #'magit-apply-hunk args))
 
 (defun magit-reverse-hunks (sections args)
   (magit-confirm 'reverse
-    (format "Reverse %s hunks from %s"
-            (length sections)
-            (magit-section-parent-value (car sections))))
-  (magit-reverse-apply sections 'magit-apply-hunks args))
+    (list "Reverse %d hunks from %s"
+          (length sections)
+          (magit-section-parent-value (car sections))))
+  (magit-reverse-apply sections #'magit-apply-hunks args))
 
 (defun magit-reverse-file (section args)
   (magit-reverse-files (list section) args))
@@ -712,10 +759,10 @@ so causes the change to be applied to the index as well."
                  (--separate (member (oref it value) bs)
                              sections))))
     (magit-confirm-files 'reverse (--map (oref it value) sections))
-    (cond ((= (length sections) 1)
-           (magit-reverse-apply (car sections) 'magit-apply-diff args))
+    (cond ((length= sections 1)
+           (magit-reverse-apply (car sections) #'magit-apply-diff args))
           (sections
-           (magit-reverse-apply sections 'magit-apply-diffs args)))
+           (magit-reverse-apply sections #'magit-apply-diffs args)))
     (when binaries
       (user-error "Cannot reverse binary files"))))
 
@@ -736,7 +783,7 @@ a separate commit.  A typical workflow would be:
 1. Visit the `HEAD' commit and navigate to the change that should
    not have been included in that commit.
 2. Type \"u\" (`magit-unstage') to reverse it in the index.
-   This assumes that `magit-unstage-committed-changes' is non-nil.
+   This assumes that `magit-unstage-committed' is non-nil.
 3. Type \"c e\" to extend `HEAD' with the staged changes,
    including those that were already staged before.
 4. Optionally stage the remaining changes using \"s\" or \"S\"
@@ -766,6 +813,11 @@ a separate commit.  A typical workflow would be:
   (interactive)
   (magit-call-smerge #'smerge-keep-lower))
 
+(defun magit-smerge-keep-all ()
+  "Keep all versions of the conflict at point."
+  (interactive)
+  (magit-call-smerge #'smerge-keep-all))
+
 (defun magit-call-smerge (fn)
   (pcase-let* ((file (magit-file-at-point t t))
                (keep (get-file-buffer file))
@@ -781,8 +833,8 @@ a separate commit.  A typical workflow would be:
           (condition-case nil
               (smerge-match-conflict)
             (error
-             (if (eq fn 'smerge-keep-current)
-                 (when (eq this-command 'magit-discard)
+             (if (eq fn #'smerge-keep-current)
+                 (when (eq this-command #'magit-discard)
                    (re-search-forward smerge-begin-re nil t)
                    (setq fn
                          (magit-read-char-case "Keep side: " t
