@@ -1,19 +1,16 @@
-;;; magit-stash.el --- stash support for Magit  -*- lexical-binding: t -*-
+;;; magit-stash.el --- Stash support for Magit  -*- lexical-binding:t -*-
 
-;; Copyright (C) 2008-2021  The Magit Project Contributors
-;;
-;; You should have received a copy of the AUTHORS.md file which
-;; lists all contributors.  If not, see http://magit.vc/authors.
+;; Copyright (C) 2008-2024 The Magit Project Contributors
 
-;; Author: Jonas Bernoulli <jonas@bernoul.li>
-;; Maintainer: Jonas Bernoulli <jonas@bernoul.li>
+;; Author: Jonas Bernoulli <emacs.magit@jonas.bernoulli.dev>
+;; Maintainer: Jonas Bernoulli <emacs.magit@jonas.bernoulli.dev>
 
 ;; SPDX-License-Identifier: GPL-3.0-or-later
 
-;; Magit is free software; you can redistribute it and/or modify it
+;; Magit is free software: you can redistribute it and/or modify it
 ;; under the terms of the GNU General Public License as published by
-;; the Free Software Foundation; either version 3, or (at your option)
-;; any later version.
+;; the Free Software Foundation, either version 3 of the License, or
+;; (at your option) any later version.
 ;;
 ;; Magit is distributed in the hope that it will be useful, but WITHOUT
 ;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
@@ -21,7 +18,7 @@
 ;; License for more details.
 ;;
 ;; You should have received a copy of the GNU General Public License
-;; along with Magit.  If not, see http://www.gnu.org/licenses.
+;; along with Magit.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Commentary:
 
@@ -31,9 +28,7 @@
 
 (require 'magit)
 (require 'magit-reflog)
-
-;; For `magit-stash-drop'.
-(defvar helm-comp-read-use-marked)
+(require 'magit-sequence)
 
 ;;; Options
 
@@ -81,10 +76,15 @@ AUTHOR-WIDTH has to be an integer.  When the name of the author
   :group 'magit-stash
   :group 'magit-margin
   :type magit-log-margin--custom-type
-  :initialize 'magit-custom-initialize-reset
+  :initialize #'magit-custom-initialize-reset
   :set-after '(magit-log-margin)
   :set (apply-partially #'magit-margin-set-variable 'magit-stashes-mode))
 
+;;;; Variables
+
+(defvar magit-stash-read-message-function #'magit-stash-read-message
+  "Function used to read the message when creating a stash.")
+
 ;;; Commands
 
 ;;;###autoload (autoload 'magit-stash "magit-stash" nil t)
@@ -98,7 +98,8 @@ AUTHOR-WIDTH has to be an integer.  When the name of the author
     ("z" "both"          magit-stash-both)
     ("i" "index"         magit-stash-index)
     ("w" "worktree"      magit-stash-worktree)
-    ("x" "keeping index" magit-stash-keep-index)]
+    ("x" "keeping index" magit-stash-keep-index)
+    ("P" "push"          magit-stash-push :level 5)]
    ["Snapshot"
     ("Z" "both"          magit-snapshot-both)
     ("I" "index"         magit-snapshot-index)
@@ -164,16 +165,29 @@ while two prefix arguments are equivalent to `--all'."
   (magit-stash-save message t t include-untracked t 'index))
 
 (defun magit-stash-read-args ()
-  (list (magit-stash-read-message)
+  (list (funcall magit-stash-read-message-function)
         (magit-stash-read-untracked)))
 
-(defun magit-stash-read-untracked ()
-  (let ((prefix (prefix-numeric-value current-prefix-arg))
-        (args   (magit-stash-arguments)))
-    (cond ((or (= prefix 16) (member "--all" args)) 'all)
-          ((or (= prefix  4) (member "--include-untracked" args)) t))))
-
 (defun magit-stash-read-message ()
+  "Read a message from the minibuffer, to be used for a stash.
+
+The message that Git would have picked, is available as the
+default (used when the user enters the empty string) and as
+the next history element (which can be accessed with \
+\\<minibuffer-local-map>\\[next-history-element])."
+  (read-string (format "Stash message (default: On%s:%s): "
+                       (magit--ellipsis) (magit--ellipsis))
+               nil nil
+               (format "On %s: %s"
+                       (or (magit-get-current-branch) "(no branch)")
+                       (magit-rev-format "%h %s"))))
+
+(defun magit-stash-read-message-traditional ()
+  "Read a message from the minibuffer, to be used for a stash.
+
+If the user confirms the initial-input unmodified, then the
+abbreviated commit hash and commit summary are appended.
+The resulting message is what Git would have used."
   (let* ((default (format "On %s: "
                           (or (magit-get-current-branch) "(no branch)")))
          (input (magit-read-string "Stash message" default)))
@@ -181,6 +195,12 @@ while two prefix arguments are equivalent to `--all'."
         (concat default (magit-rev-format "%h %s"))
       input)))
 
+(defun magit-stash-read-untracked ()
+  (let ((prefix (prefix-numeric-value current-prefix-arg))
+        (args   (magit-stash-arguments)))
+    (cond ((or (= prefix 16) (member "--all" args)) 'all)
+          ((or (= prefix  4) (member "--include-untracked" args)) t))))
+
 ;;;###autoload
 (defun magit-snapshot-both (&optional include-untracked)
   "Create a snapshot of the index and working tree.
@@ -213,36 +233,108 @@ while two prefix arguments are equivalent to `--all'."
   (magit-stash-save (concat "WIP on " (magit-stash-summary))
                     index worktree untracked refresh t))
 
+;;;###autoload (autoload 'magit-stash-push "magit-stash" nil t)
+(transient-define-prefix magit-stash-push (&optional transient args)
+  "Create stash using \"git stash push\".
+
+This differs from Magit's other stashing commands, which don't
+use \"git stash\" and are generally more flexible but don't allow
+specifying a list of files to be stashed."
+  :man-page "git-stash"
+  ["Arguments"
+   (magit:-- :reader (lambda (prompt initial-input history)
+                       (magit-read-files prompt initial-input history
+                                         #'magit-modified-files)))
+   ("-u" "Also save untracked files" ("-u" "--include-untracked"))
+   ("-a" "Also save untracked and ignored files" ("-a" "--all"))
+   ("-k" "Keep index" ("-k" "--keep-index"))
+   ("-K" "Don't keep index" "--no-keep-index")]
+  ["Actions"
+   ("P" "push" magit-stash-push)]
+  (interactive (if (eq transient-current-command 'magit-stash-push)
+                   (list nil (transient-args 'magit-stash-push))
+                 (list t)))
+  (if transient
+      (transient-setup 'magit-stash-push)
+    (magit-run-git "stash" "push" args)))
+
 ;;;###autoload
 (defun magit-stash-apply (stash)
   "Apply a stash to the working tree.
-Try to preserve the stash index.  If that fails because there
-are staged changes, apply without preserving the stash index."
+
+First try \"git stash apply --index\", which tries to preserve
+the index stored in the stash, if any.  This may fail because
+applying the stash could result in conflicts and those have to
+be stored in the index, making it impossible to also store the
+stash's index there as well.
+
+If the above failed, then try \"git stash apply\".  This fails
+\(with or without \"--index\") if there are any uncommitted
+changes to files that are also modified in the stash.
+
+If both of the above failed, then apply using \"git apply\".
+If there are no conflicting files, use \"--3way\".  If there are
+conflicting files, then using \"--3way\" requires that those
+files are staged first, which may be undesirable, so prompt
+the user whether to use \"--3way\" or \"--reject\"."
   (interactive (list (magit-read-stash "Apply stash")))
-  (if (= (magit-call-git "stash" "apply" "--index" stash) 0)
-      (magit-refresh)
-    (magit-run-git "stash" "apply" stash)))
+  (magit-stash--apply "apply" stash))
 
 ;;;###autoload
 (defun magit-stash-pop (stash)
-  "Apply a stash to the working tree and remove it from stash list.
-Try to preserve the stash index.  If that fails because there
-are staged changes, apply without preserving the stash index
-and forgo removing the stash."
+  "Apply a stash to the working tree, on success remove it from stash list.
+
+First try \"git stash pop --index\", which tries to preserve
+the index stored in the stash, if any.  This may fail because
+applying the stash could result in conflicts and those have to
+be stored in the index, making it impossible to also store the
+stash's index there as well.
+
+If the above failed, then try \"git stash apply\".  This fails
+\(with or without \"--index\") if there are any uncommitted
+changes to files that are also modified in the stash.
+
+If both of the above failed, then apply using \"git apply\".
+If there are no conflicting files, use \"--3way\".  If there are
+conflicting files, then using \"--3way\" requires that those
+files are staged first, which may be undesirable, so prompt
+the user whether to use \"--3way\" or \"--reject\"."
   (interactive (list (magit-read-stash "Pop stash")))
-  (if (= (magit-call-git "stash" "apply" "--index" stash) 0)
-      (magit-stash-drop stash)
-    (magit-run-git "stash" "apply" stash)))
+  (magit-stash--apply "pop" stash))
+
+(defun magit-stash--apply (action stash)
+  (or (= (magit-call-git "stash" action "--index" stash) 0)
+      ;; The stash's index could not be applied, so always keep the stash.
+      (= (magit-call-git "stash" "apply" stash) 0)
+      (let* ((range (format "%s^..%s" stash stash))
+             (stashed (magit-git-items "diff" "-z" "--name-only" range "--"))
+             (conflicts (cl-sort (cl-union (magit-unstaged-files t stashed)
+                                           (magit-untracked-files t stashed)
+                                           :test #'equal)
+                                 #'string<))
+             (arg (cond
+                   ((not conflicts) "--3way")
+                   ((magit-confirm-files
+                     'stash-apply-3way conflicts
+                     "Apply stash using `--3way', which requires first staging"
+                     "(else use `--reject')"
+                     t)
+                    (magit-stage-1 nil conflicts)
+                    "--3way")
+                   ("--reject"))))
+        (with-temp-buffer
+          (magit-git-insert "diff" range)
+          (magit-run-git-with-input "apply" arg "-"))))
+  (magit-refresh))
 
 ;;;###autoload
 (defun magit-stash-drop (stash)
   "Remove a stash from the stash list.
 When the region is active offer to drop all contained stashes."
   (interactive
-   (list (--if-let (magit-region-values 'stash)
-             (magit-confirm 'drop-stashes nil "Drop %i stashes" nil it)
-           (let ((helm-comp-read-use-marked t))
-             (magit-read-stash "Drop stash")))))
+   (list (if-let ((values (magit-region-values 'stash)))
+             (magit-confirm 'drop-stashes nil "Drop %d stashes" nil values)
+           (magit-read-stash "Drop stash"))))
   (dolist (stash (if (listp stash)
                      (nreverse (prog1 stash (setq stash (car stash))))
                    (list stash)))
@@ -256,26 +348,29 @@ When the region is active offer to drop all contained stashes."
 (defun magit-stash-clear (ref)
   "Remove all stashes saved in REF's reflog by deleting REF."
   (interactive (let ((ref (or (magit-section-value-if 'stashes) "refs/stash")))
-                 (magit-confirm t (format "Drop all stashes in %s" ref))
+                 (magit-confirm t (list "Drop all stashes in %s" ref))
                  (list ref)))
   (magit-run-git "update-ref" "-d" ref))
 
 ;;;###autoload
 (defun magit-stash-branch (stash branch)
-  "Create and checkout a new BRANCH from STASH."
+  "Create and checkout a new BRANCH from an existing STASH.
+The new branch starts at the commit that was current when the
+stash was created.  If the stash applies cleanly, then drop it."
   (interactive (list (magit-read-stash "Branch stash")
                      (magit-read-string-ns "Branch name")))
   (magit-run-git "stash" "branch" branch stash))
 
 ;;;###autoload
 (defun magit-stash-branch-here (stash branch)
-  "Create and checkout a new BRANCH and apply STASH.
-The branch is created using `magit-branch-and-checkout', using the
-current branch or `HEAD' as the start-point."
+  "Create and checkout a new BRANCH from an existing STASH.
+Use the current branch or `HEAD' as the starting-point of BRANCH.
+Then apply STASH, dropping it if it applies cleanly."
   (interactive (list (magit-read-stash "Branch stash")
                      (magit-read-string-ns "Branch name")))
-  (let ((magit-inhibit-refresh t))
-    (magit-branch-and-checkout branch (or (magit-get-current-branch) "HEAD")))
+  (let ((start-point (or (magit-get-current-branch) "HEAD")))
+    (magit-call-git "checkout" "-b" branch start-point)
+    (magit-branch-maybe-adjust-upstream branch start-point))
   (magit-stash-apply stash))
 
 ;;;###autoload
@@ -298,7 +393,7 @@ current branch or `HEAD' as the start-point."
                            (magit-stash-create message index worktree untracked))
         (if (eq keep 'worktree)
             (with-temp-buffer
-              (magit-git-insert "diff" "--cached")
+              (magit-git-insert "diff" "--cached" "--no-ext-diff")
               (magit-run-git-with-input
                "apply" "--reverse" "--cached" "--ignore-space-change" "-")
               (magit-run-git-with-input
@@ -353,23 +448,26 @@ current branch or `HEAD' as the start-point."
 
 ;;; Sections
 
-(defvar magit-stashes-section-map
-  (let ((map (make-sparse-keymap)))
-    (define-key map [remap magit-delete-thing] 'magit-stash-clear)
-    map)
-  "Keymap for `stashes' section.")
+(defvar-keymap magit-stashes-section-map
+  :doc "Keymap for `stashes' section."
+  "<remap> <magit-delete-thing>" #'magit-stash-clear
+  "<remap> <magit-visit-thing>"  #'magit-stash-list
+  "<2>" (magit-menu-item "Clear %t" #'magit-stash-clear)
+  "<1>" (magit-menu-item "List %t"  #'magit-stash-list))
 
-(defvar magit-stash-section-map
-  (let ((map (make-sparse-keymap)))
-    (define-key map [remap magit-visit-thing]  'magit-stash-show)
-    (define-key map [remap magit-delete-thing] 'magit-stash-drop)
-    (define-key map "a"  'magit-stash-apply)
-    (define-key map "A"  'magit-stash-pop)
-    map)
-  "Keymap for `stash' sections.")
+(defvar-keymap magit-stash-section-map
+  :doc "Keymap for `stash' sections."
+  "<remap> <magit-cherry-pick>"  #'magit-stash-pop
+  "<remap> <magit-cherry-apply>" #'magit-stash-apply
+  "<remap> <magit-delete-thing>" #'magit-stash-drop
+  "<remap> <magit-visit-thing>"  #'magit-stash-show
+  "<4>" (magit-menu-item "Pop %M"    #'magit-stash-pop)
+  "<3>" (magit-menu-item "Apply %M"  #'magit-stash-apply)
+  "<2>" (magit-menu-item "Delete %M" #'magit-stash-drop)
+  "<1>" (magit-menu-item "Visit %v"  #'magit-stash-show))
 
 (magit-define-section-jumper magit-jump-to-stashes
-  "Stashes" stashes "refs/stash")
+  "Stashes" stashes "refs/stash" magit-insert-stashes)
 
 (cl-defun magit-insert-stashes (&optional (ref   "refs/stash")
                                           (heading "Stashes:"))
@@ -378,14 +476,7 @@ If optional REF is non-nil, show reflog for that instead.
 If optional HEADING is non-nil, use that as section heading
 instead of \"Stashes:\"."
   (let ((verified (magit-rev-verify ref))
-        (autostash
-         (and (magit-rebase-in-progress-p)
-              (thread-first
-                  (if (file-directory-p (magit-git-dir "rebase-merge"))
-                      "rebase-merge/autostash"
-                    "rebase-apply/autostash")
-                magit-git-dir
-                magit-file-line))))
+        (autostash (magit-rebase--get-state-lines "autostash")))
     (when (or autostash verified)
       (magit-insert-section (stashes ref)
         (magit-insert-heading heading)
@@ -403,7 +494,7 @@ instead of \"Stashes:\"."
                 (backward-char)
                 (magit-log-format-margin autostash author date)))))
         (if verified
-            (magit-git-wash (apply-partially 'magit-log-wash-log 'stash)
+            (magit-git-wash (apply-partially #'magit-log-wash-log 'stash)
               "reflog" "--format=%gd%x00%aN%x00%at%x00%gs" ref)
           (insert ?\n)
           (save-excursion
@@ -420,8 +511,9 @@ instead of \"Stashes:\"."
 
 (define-derived-mode magit-stashes-mode magit-reflog-mode "Magit Stashes"
   "Mode for looking at lists of stashes."
+  :interactive nil
   :group 'magit-log
-  (hack-dir-local-variables-non-file-buffer))
+  (magit-hack-dir-local-variables))
 
 (defun magit-stashes-setup-buffer ()
   (magit-setup-buffer #'magit-stashes-mode nil
@@ -429,10 +521,11 @@ instead of \"Stashes:\"."
 
 (defun magit-stashes-refresh-buffer ()
   (magit-insert-section (stashesbuf)
-    (magit-insert-heading (if (equal magit-buffer-refname "refs/stash")
-                              "Stashes:"
-                            (format "Stashes [%s]:" magit-buffer-refname)))
-    (magit-git-wash (apply-partially 'magit-log-wash-log 'stash)
+    (magit-insert-heading t
+      (if (equal magit-buffer-refname "refs/stash")
+          "Stashes"
+        (format "Stashes [%s]" magit-buffer-refname)))
+    (magit-git-wash (apply-partially #'magit-log-wash-log 'stash)
       "reflog" "--format=%gd%x00%aN%x00%at%x00%gs" magit-buffer-refname)))
 
 (cl-defmethod magit-buffer-value (&context (major-mode magit-stashes-mode))
@@ -447,8 +540,8 @@ If there is no stash buffer in the same frame, then do nothing."
     (magit--maybe-update-stash-buffer)))
 
 (defun magit--maybe-update-stash-buffer ()
-  (when-let ((stash  (magit-section-value-if 'stash))
-             (buffer (magit-get-mode-buffer 'magit-stash-mode nil t)))
+  (when-let* ((stash  (magit-section-value-if 'stash))
+              (buffer (magit-get-mode-buffer 'magit-stash-mode nil t)))
     (if magit--update-stash-buffer
         (setq magit--update-stash-buffer (list stash buffer))
       (setq magit--update-stash-buffer (list stash buffer))
@@ -480,8 +573,10 @@ If there is no stash buffer in the same frame, then do nothing."
 
 (define-derived-mode magit-stash-mode magit-diff-mode "Magit Stash"
   "Mode for looking at individual stashes."
+  :interactive nil
   :group 'magit-diff
-  (hack-dir-local-variables-non-file-buffer))
+  (magit-hack-dir-local-variables)
+  (setq magit--imenu-group-types '(commit)))
 
 (defun magit-stash-setup-buffer (stash args files)
   (magit-setup-buffer #'magit-stash-mode nil
@@ -507,20 +602,19 @@ If there is no stash buffer in the same frame, then do nothing."
 (defun magit-stash-insert-section (commit range message &optional files)
   (magit-insert-section (commit commit)
     (magit-insert-heading message)
-    (magit--insert-diff "diff" range "-p" "--no-prefix" magit-buffer-diff-args
-                        "--" (or files magit-buffer-diff-files))))
+    (magit--insert-diff nil
+      "diff" range "-p" "--no-prefix" magit-buffer-diff-args
+      "--" (or files magit-buffer-diff-files))))
 
 (defun magit-insert-stash-notes ()
   "Insert section showing notes for a stash.
 This shows the notes for stash@{N} but not for the other commits
 that make up the stash."
-  (magit-insert-section section (note)
-    (magit-insert-heading "Notes")
+  (magit-insert-section (note)
+    (magit-insert-heading t "Notes")
     (magit-git-insert "notes" "show" magit-buffer-revision)
-    (if (= (point)
-           (oref section content))
-        (magit-cancel-section)
-      (insert "\n"))))
+    (magit-cancel-section 'if-empty)
+    (insert "\n")))
 
 (defun magit-insert-stash-index ()
   "Insert section showing staged changes of the stash."
