@@ -1,19 +1,16 @@
-;;; magit-status.el --- the grand overview  -*- lexical-binding: t -*-
+;;; magit-status.el --- The grand overview  -*- lexical-binding:t -*-
 
-;; Copyright (C) 2010-2021  The Magit Project Contributors
-;;
-;; You should have received a copy of the AUTHORS.md file which
-;; lists all contributors.  If not, see http://magit.vc/authors.
+;; Copyright (C) 2008-2024 The Magit Project Contributors
 
-;; Author: Jonas Bernoulli <jonas@bernoul.li>
-;; Maintainer: Jonas Bernoulli <jonas@bernoul.li>
+;; Author: Jonas Bernoulli <emacs.magit@jonas.bernoulli.dev>
+;; Maintainer: Jonas Bernoulli <emacs.magit@jonas.bernoulli.dev>
 
 ;; SPDX-License-Identifier: GPL-3.0-or-later
 
-;; Magit is free software; you can redistribute it and/or modify it
+;; Magit is free software: you can redistribute it and/or modify it
 ;; under the terms of the GNU General Public License as published by
-;; the Free Software Foundation; either version 3, or (at your option)
-;; any later version.
+;; the Free Software Foundation, either version 3 of the License, or
+;; (at your option) any later version.
 ;;
 ;; Magit is distributed in the hope that it will be useful, but WITHOUT
 ;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
@@ -21,7 +18,7 @@
 ;; License for more details.
 ;;
 ;; You should have received a copy of the GNU General Public License
-;; along with Magit.  If not, see http://www.gnu.org/licenses.
+;; along with Magit.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Commentary:
 
@@ -171,7 +168,7 @@ AUTHOR-WIDTH has to be an integer.  When the name of the author
   :group 'magit-status
   :group 'magit-margin
   :type magit-log-margin--custom-type
-  :initialize 'magit-custom-initialize-reset
+  :initialize #'magit-custom-initialize-reset
   :set-after '(magit-log-margin)
   :set (apply-partially #'magit-margin-set-variable 'magit-status-mode))
 
@@ -292,8 +289,10 @@ prefix arguments:
 (put 'magit-status 'interactive-only 'magit-status-setup-buffer)
 
 ;;;###autoload
-(defalias 'magit 'magit-status
-  "An alias for `magit-status' for better discoverability.
+(defalias 'magit #'magit-status
+  "Begin using Magit.
+
+This alias for `magit-status' exists for better discoverability.
 
 Instead of invoking this alias for `magit-status' using
 \"M-x magit RET\", you should bind a key to `magit-status'
@@ -309,6 +308,7 @@ also contains other useful hints.")
 
 (put 'magit-status-here 'interactive-only 'magit-status-setup-buffer)
 
+;;;###autoload
 (defun magit-status-quick ()
   "Show the status of the current Git repository, maybe without refreshing.
 
@@ -323,7 +323,7 @@ Prefix arguments have the same meaning as for `magit-status',
 and additionally cause the buffer to be refresh.
 
 To use this function instead of `magit-status', add this to your
-init file: (global-set-key (kbd \"C-x g\") 'magit-status-quick)."
+init file: (global-set-key (kbd \"C-x g\") \\='magit-status-quick)."
   (interactive)
   (if-let ((buffer
             (and (not current-prefix-arg)
@@ -332,74 +332,32 @@ init file: (global-set-key (kbd \"C-x g\") 'magit-status-quick)."
       (magit-display-buffer buffer)
     (call-interactively #'magit-status)))
 
-(defvar magit--remotes-using-recent-git nil)
-
-(defun magit--tramp-asserts (directory)
-  (when-let ((remote (file-remote-p directory)))
-    (unless (member remote magit--remotes-using-recent-git)
-      (if-let ((version (let ((default-directory directory))
-                          (magit-git-version))))
-          (if (version<= magit--minimal-git version)
-              (push remote magit--remotes-using-recent-git)
-            (display-warning 'magit (format "\
-Magit requires Git >= %s, but on %s the version is %s.
-
-If multiple Git versions are installed on the host, then the
-problem might be that TRAMP uses the wrong executable.
-
-Check the value of `magit-remote-git-executable' and consult
-the info node `(tramp)Remote programs'.
-" magit--minimal-git remote version) :error))
-        (display-warning 'magit (format "\
-Magit cannot find Git on %s.
-
-Check the value of `magit-remote-git-executable' and consult
-the info node `(tramp)Remote programs'." remote) :error)))))
-
 ;;; Mode
 
-(defvar magit-status-mode-map
-  (let ((map (make-sparse-keymap)))
-    (set-keymap-parent map magit-mode-map)
-    (define-key map "j" 'magit-status-jump)
-    (define-key map [remap dired-jump] 'magit-dired-jump)
-    map)
-  "Keymap for `magit-status-mode'.")
+(defvar-keymap magit-status-mode-map
+  :doc "Keymap for `magit-status-mode'."
+  :parent magit-mode-map
+  "j" #'magit-status-jump
+  "<remap> <dired-jump>" #'magit-dired-jump)
 
 (transient-define-prefix magit-status-jump ()
   "In a Magit-Status buffer, jump to a section."
-  ["Jump to"
-   [("z " "Stashes" magit-jump-to-stashes
-     :if (lambda () (memq 'magit-insert-stashes magit-status-sections-hook)))
-    ("t " "Tracked" magit-jump-to-tracked
-     :if (lambda () (memq 'magit-insert-tracked-files magit-status-sections-hook)))
-    ("n " "Untracked" magit-jump-to-untracked
-     :if (lambda () (memq 'magit-insert-untracked-files magit-status-sections-hook)))
-    ("u " "Unstaged" magit-jump-to-unstaged
-     :if (lambda () (memq 'magit-insert-unstaged-changes magit-status-sections-hook)))
-    ("s " "Staged" magit-jump-to-staged
-     :if (lambda () (memq 'magit-insert-staged-changes magit-status-sections-hook)))]
-   [("fu" "Unpulled from upstream" magit-jump-to-unpulled-from-upstream
-     :if (lambda () (memq 'magit-insert-unpulled-from-upstream magit-status-sections-hook)))
-    ("fp" "Unpulled from pushremote" magit-jump-to-unpulled-from-pushremote
-     :if (lambda () (memq 'magit-insert-unpulled-from-pushremote magit-status-sections-hook)))
-    ("pu" magit-jump-to-unpushed-to-upstream
-     :if (lambda ()
-           (or (memq 'magit-insert-unpushed-to-upstream-or-recent magit-status-sections-hook)
-               (memq 'magit-insert-unpushed-to-upstream magit-status-sections-hook)))
-     :description (lambda ()
-                    (let ((upstream (magit-get-upstream-branch)))
-                      (if (or (not upstream)
-                              (magit-rev-ancestor-p "HEAD" upstream))
-                          "Recent commits"
-                        "Unmerged into upstream"))))
-    ("pp" "Unpushed to pushremote" magit-jump-to-unpushed-to-pushremote
-     :if (lambda () (memq 'magit-insert-unpushed-to-pushremote magit-status-sections-hook)))
-    ("a " "Assumed unstaged" magit-jump-to-assume-unchanged
-     :if (lambda () (memq 'magit-insert-assume-unchanged-files magit-status-sections-hook)))
-    ("w " "Skip worktree" magit-jump-to-skip-worktree
-     :if (lambda () (memq 'magit-insert-skip-worktree-files magit-status-sections-hook)))]
-   [("i" "Using Imenu" imenu)]])
+  [["Jump to"
+    ("z " magit-jump-to-stashes)
+    ("t " magit-jump-to-tracked)
+    ("n " magit-jump-to-untracked)
+    ("i " magit-jump-to-ignored)
+    ("u " magit-jump-to-unstaged)
+    ("s " magit-jump-to-staged)]
+   [""
+    ("fu" magit-jump-to-unpulled-from-upstream)
+    ("fp" magit-jump-to-unpulled-from-pushremote)
+    ("pu" magit-jump-to-unpushed-to-upstream)
+    ("pp" magit-jump-to-unpushed-to-pushremote)
+    ("a " magit-jump-to-assume-unchanged)
+    ("w " magit-jump-to-skip-worktree)]
+   ["Jump using"
+    ("j"  "Imenu" imenu)]])
 
 (define-derived-mode magit-status-mode magit-mode "Magit"
   "Mode for looking at Git status.
@@ -427,10 +385,13 @@ Staging and applying changes is documented in info node
 Type \\[magit-commit] to create a commit.
 
 \\{magit-status-mode-map}"
+  :interactive nil
   :group 'magit-status
-  (hack-dir-local-variables-non-file-buffer)
-  (setq imenu-create-index-function
-        'magit-imenu--status-create-index-function))
+  (magit-hack-dir-local-variables)
+  (when magit-status-initial-section
+    (add-hook 'magit-post-create-buffer-hook
+              #'magit-status-goto-initial-section nil t))
+  (setq magit--imenu-group-types '(not branch commit)))
 
 (put 'magit-status-mode 'magit-diff-default-arguments
      '("--no-ext-diff"))
@@ -441,7 +402,8 @@ Type \\[magit-commit] to create a commit.
 (defun magit-status-setup-buffer (&optional directory)
   (unless directory
     (setq directory default-directory))
-  (magit--tramp-asserts directory)
+  (when (file-remote-p directory)
+    (magit-git-version-assert))
   (let* ((default-directory directory)
          (d (magit-diff--get-value 'magit-status-mode
                                    magit-status-use-buffer-arguments))
@@ -449,8 +411,8 @@ Type \\[magit-commit] to create a commit.
                                   magit-status-use-buffer-arguments))
          (file (and magit-status-goto-file-position
                     (magit-file-relative-name)))
-         (line (and file (line-number-at-pos)))
-         (col  (and file (current-column)))
+         (line (and file (save-restriction (widen) (line-number-at-pos))))
+         (col  (and file (save-restriction (widen) (current-column))))
          (buf  (magit-setup-buffer #'magit-status-mode nil
                  (magit-buffer-diff-args  (nth 0 d))
                  (magit-buffer-diff-files (nth 1 d))
@@ -475,19 +437,7 @@ Type \\[magit-commit] to create a commit.
     (magit-run-section-hook 'magit-status-sections-hook)))
 
 (defun magit-status-goto-initial-section ()
-  "In a `magit-status-mode' buffer, jump `magit-status-initial-section'.
-Actually doing so is deferred until `magit-refresh-buffer-hook'
-runs `magit-status-goto-initial-section-1'.  That function then
-removes itself from the hook, so that this only happens when the
-status buffer is first created."
-  (when (and magit-status-initial-section
-             (derived-mode-p 'magit-status-mode))
-    (add-hook 'magit-refresh-buffer-hook
-              'magit-status-goto-initial-section-1 nil t)))
-
-(defun magit-status-goto-initial-section-1 ()
-  "In a `magit-status-mode' buffer, jump `magit-status-initial-section'.
-This function removes itself from `magit-refresh-buffer-hook'."
+  "Jump to the section specified by `magit-status-initial-section'."
   (when-let ((section
               (--some (if (integerp it)
                           (nth (1- it)
@@ -500,9 +450,7 @@ This function removes itself from `magit-refresh-buffer-hook'."
                                magit-section-initial-visibility-alist))))
       (if (eq vis 'hide)
           (magit-section-hide section)
-        (magit-section-show section))))
-  (remove-hook 'magit-refresh-buffer-hook
-               'magit-status-goto-initial-section-1 t))
+        (magit-section-show section)))))
 
 (defun magit-status-maybe-update-revision-buffer (&optional _)
   "When moving in the status buffer, update the revision buffer.
@@ -533,11 +481,10 @@ The sections are inserted by running the functions on the hook
       (magit-insert-headers 'magit-status-headers-hook)
     (insert "In the beginning there was darkness\n\n")))
 
-(defvar magit-error-section-map
-  (let ((map (make-sparse-keymap)))
-    (define-key map [remap magit-visit-thing] 'magit-process-buffer)
-    map)
-  "Keymap for `error' sections.")
+(defvar-keymap magit-error-section-map
+  :doc "Keymap for `error' sections."
+  "<remap> <magit-visit-thing>" #'magit-process-buffer
+  "<1>" (magit-menu-item "Visit process output" #'magit-process-buffer))
 
 (defun magit-insert-error-header ()
   "Insert the message about the Git error that just occurred.
@@ -550,8 +497,7 @@ the status buffer causes this section to disappear again."
     (magit-insert-section (error 'git)
       (insert (propertize (format "%-10s" "GitError! ")
                           'font-lock-face 'magit-section-heading))
-      (insert (propertize magit-this-error
-                          'font-lock-face 'font-lock-warning-face))
+      (insert (propertize magit-this-error 'font-lock-face 'error))
       (when-let ((key (car (where-is-internal 'magit-process-buffer))))
         (insert (format "  [Type `%s' for details]" (key-description key))))
       (insert ?\n))
@@ -569,7 +515,7 @@ the status buffer causes this section to disappear again."
         (when magit-buffer-diff-files
           (insert " -- ")))
       (when magit-buffer-diff-files
-        (insert (mapconcat #'identity magit-buffer-diff-files " ")))
+        (insert (string-join magit-buffer-diff-files " ")))
       (insert ?\n))))
 
 ;;;; Reference Headers
@@ -638,31 +584,31 @@ arguments are for internal use only."
                    (concat
                     (propertize merge 'font-lock-face 'magit-branch-local) " "
                     (propertize "does not exist"
-                                'font-lock-face 'font-lock-warning-face))
+                                'font-lock-face 'magit-branch-warning))
                  (format
                   "%s %s %s"
                   (propertize merge 'font-lock-face 'magit-branch-remote)
                   (propertize "does not exist on"
-                              'font-lock-face 'font-lock-warning-face)
+                              'font-lock-face 'magit-branch-warning)
                   (propertize remote 'font-lock-face 'magit-branch-remote))))
               (t
                (propertize "invalid upstream configuration"
-                           'font-lock-face 'font-lock-warning-face)))))
+                           'font-lock-face 'magit-branch-warning)))))
           (insert ?\n))))))
 
 (defun magit-insert-push-branch-header ()
   "Insert a header line about the branch the current branch is pushed to."
-  (when-let ((branch (magit-get-current-branch))
-             (target (magit-get-push-branch branch)))
+  (when-let* ((branch (magit-get-current-branch))
+              (target (magit-get-push-branch branch)))
     (magit-insert-section (branch target)
       (insert (format "%-10s" "Push: "))
       (insert
        (if (magit-rev-verify target)
-           (concat target " "
-                   (and magit-status-show-hashes-in-headers
+           (concat (and magit-status-show-hashes-in-headers
                         (concat (propertize (magit-rev-format "%h" target)
                                             'font-lock-face 'magit-hash)
                                 " "))
+                   target " "
                    (funcall magit-log-format-message-function target
                             (funcall magit-log-format-message-function nil
                                      (or (magit-rev-format "%s" target)
@@ -671,10 +617,10 @@ arguments are for internal use only."
            (if (magit-remote-p remote)
                (concat target " "
                        (propertize "does not exist"
-                                   'font-lock-face 'font-lock-warning-face))
+                                   'font-lock-face 'magit-branch-warning))
              (concat remote " "
                      (propertize "remote does not exist"
-                                 'font-lock-face 'font-lock-warning-face))))))
+                                 'font-lock-face 'magit-branch-warning))))))
       (insert ?\n))))
 
 (defun magit-insert-tags-header ()
@@ -689,14 +635,13 @@ arguments are for internal use only."
     (when (or this-tag next-tag)
       (magit-insert-section (tag (or this-tag next-tag))
         (insert (format "%-10s" (if both-tags "Tags: " "Tag: ")))
-        (cl-flet ((insert-count
-                   (tag count face)
-                   (insert (concat (propertize tag 'font-lock-face 'magit-tag)
-                                   (and (> count 0)
-                                        (format " (%s)"
-                                                (propertize
-                                                 (format "%s" count)
-                                                 'font-lock-face face)))))))
+        (cl-flet ((insert-count (tag count face)
+                    (insert (concat (propertize tag 'font-lock-face 'magit-tag)
+                                    (and (> count 0)
+                                         (format " (%s)"
+                                                 (propertize
+                                                  (format "%s" count)
+                                                  'font-lock-face face)))))))
           (when this-tag  (insert-count this-tag this-cnt 'magit-branch-local))
           (when both-tags (insert ", "))
           (when next-tag  (insert-count next-tag next-cnt 'magit-tag)))
@@ -726,10 +671,10 @@ arguments are for internal use only."
 If no remote is configured for the current branch, then fall back
 showing the \"origin\" remote, or if that does not exist the first
 remote in alphabetic order."
-  (when-let ((name (magit-get-some-remote))
-             ;; Under certain configurations it's possible for url
-             ;; to be nil, when name is not, see #2858.
-             (url (magit-get "remote" name "url")))
+  (when-let* ((name (magit-get-some-remote))
+              ;; Under certain configurations it's possible for
+              ;; url to be nil, when name is not, see #2858.
+              (url (magit-get "remote" name "url")))
     (magit-insert-section (remote name)
       (insert (format "%-10s" "Remote: "))
       (insert (propertize name 'font-lock-face 'magit-branch-remote) ?\s)
@@ -737,14 +682,28 @@ remote in alphabetic order."
 
 ;;;; File Sections
 
-(defvar magit-untracked-section-map
-  (let ((map (make-sparse-keymap)))
-    (define-key map [remap magit-delete-thing] 'magit-discard)
-    (define-key map "s" 'magit-stage)
-    map)
-  "Keymap for the `untracked' section.")
+(defvar-keymap magit-untracked-section-map
+  :doc "Keymap for the `untracked' section."
+  "<remap> <magit-delete-thing>" #'magit-discard
+  "<remap> <magit-stage-file>"   #'magit-stage
+  "<2>" (magit-menu-item "Discard files" #'magit-discard)
+  "<1>" (magit-menu-item "Stage files"   #'magit-stage))
 
-(magit-define-section-jumper magit-jump-to-untracked "Untracked files" untracked)
+(magit-define-section-jumper magit-jump-to-untracked
+  "Untracked files" untracked nil magit-insert-untracked-files)
+
+(magit-define-section-jumper magit-jump-to-tracked
+  "Tracked files" tracked nil magit-insert-tracked-files)
+
+(magit-define-section-jumper magit-jump-to-ignored
+  "Ignored files" ignored nil magit-insert-ignored-files)
+
+(magit-define-section-jumper magit-jump-to-skip-worktree
+  "Skip-worktree files" skip-worktree nil magit-insert-skip-worktree-files)
+
+(magit-define-section-jumper magit-jump-to-assume-unchanged
+  "Assume-unchanged files" assume-unchanged nil
+  magit-insert-assume-unchanged-files)
 
 (defun magit-insert-untracked-files ()
   "Maybe insert a list or tree of untracked files.
@@ -757,30 +716,14 @@ be expanded using \"TAB\".
 If the first element of `magit-buffer-diff-files' is a
 directory, then limit the list to files below that.  The value
 value of that variable can be set using \"D -- DIRECTORY RET g\"."
-  (let* ((show (or (magit-get "status.showUntrackedFiles") "normal"))
-         (base (car magit-buffer-diff-files))
-         (base (and base (file-directory-p base) base)))
+  (let ((show (or (magit-get "status.showUntrackedFiles") "normal")))
     (unless (equal show "no")
-      (if (equal show "all")
-          (when-let ((files (magit-untracked-files nil base)))
-            (magit-insert-section (untracked)
-              (magit-insert-heading "Untracked files:")
-              (magit-insert-files files base)
-              (insert ?\n)))
-        (when-let ((files
-                    (--mapcat (and (eq (aref it 0) ??)
-                                   (list (substring it 3)))
-                              (magit-git-items "status" "-z" "--porcelain"
-                                               (magit-ignore-submodules-p t)
-                                               "--" base))))
-          (magit-insert-section (untracked)
-            (magit-insert-heading "Untracked files:")
-            (dolist (file files)
-              (magit-insert-section (file file)
-                (insert (propertize file 'font-lock-face 'magit-filename) ?\n)))
-            (insert ?\n)))))))
-
-(magit-define-section-jumper magit-jump-to-tracked "Tracked files" tracked)
+      (let* ((all (equal show "all"))
+             (base (car magit-buffer-diff-files))
+             (base (and base (file-directory-p base) base)))
+        (magit-insert-files 'untracked
+                            (lambda () (magit-untracked-files nil base (not all)))
+                            (not all))))))
 
 (defun magit-insert-tracked-files ()
   "Insert a tree of tracked files.
@@ -788,13 +731,7 @@ value of that variable can be set using \"D -- DIRECTORY RET g\"."
 If the first element of `magit-buffer-diff-files' is a
 directory, then limit the list to files below that.  The value
 value of that variable can be set using \"D -- DIRECTORY RET g\"."
-  (when-let ((files (magit-list-files)))
-    (let* ((base (car magit-buffer-diff-files))
-           (base (and base (file-directory-p base) base)))
-      (magit-insert-section (tracked nil t)
-        (magit-insert-heading "Tracked files:")
-        (magit-insert-files files base)
-        (insert ?\n)))))
+  (magit-insert-files 'tracked #'magit-list-files))
 
 (defun magit-insert-ignored-files ()
   "Insert a tree of ignored files.
@@ -802,15 +739,7 @@ value of that variable can be set using \"D -- DIRECTORY RET g\"."
 If the first element of `magit-buffer-diff-files' is a
 directory, then limit the list to files below that.  The value
 of that variable can be set using \"D -- DIRECTORY RET g\"."
-  (when-let ((files (magit-ignored-files)))
-    (let* ((base (car magit-buffer-diff-files))
-           (base (and base (file-directory-p base) base)))
-      (magit-insert-section (tracked nil t)
-        (magit-insert-heading "Ignored files:")
-        (magit-insert-files files base)
-        (insert ?\n)))))
-
-(magit-define-section-jumper magit-jump-to-skip-worktree "Skip-worktree files" skip-worktree)
+  (magit-insert-files 'ignored #'magit-ignored-files))
 
 (defun magit-insert-skip-worktree-files ()
   "Insert a tree of skip-worktree files.
@@ -818,15 +747,7 @@ of that variable can be set using \"D -- DIRECTORY RET g\"."
 If the first element of `magit-buffer-diff-files' is a
 directory, then limit the list to files below that.  The value
 of that variable can be set using \"D -- DIRECTORY RET g\"."
-  (when-let ((files (magit-skip-worktree-files)))
-    (let* ((base (car magit-buffer-diff-files))
-           (base (and base (file-directory-p base) base)))
-      (magit-insert-section (skip-worktree nil t)
-        (magit-insert-heading "Skip-worktree files:")
-        (magit-insert-files files base)
-        (insert ?\n)))))
-
-(magit-define-section-jumper magit-jump-to-assume-unchanged "Assume-unchanged files" assume-unchanged)
+  (magit-insert-files 'skip-worktree #'magit-skip-worktree-files))
 
 (defun magit-insert-assume-unchanged-files ()
   "Insert a tree of files that are assumed to be unchanged.
@@ -834,25 +755,46 @@ of that variable can be set using \"D -- DIRECTORY RET g\"."
 If the first element of `magit-buffer-diff-files' is a
 directory, then limit the list to files below that.  The value
 of that variable can be set using \"D -- DIRECTORY RET g\"."
-  (when-let ((files (magit-assume-unchanged-files)))
+  (magit-insert-files 'assume-unchanged #'magit-assume-unchanged-files))
+
+(defvar magit-file-section-indent nil
+  "Indentation used for simple `file' sections.
+If non-nil, this must be a string or character, and is used as
+indentation of `file' sections inserted by `magit-insert-*-files'.
+`file' sections that are part of diffs are not affected.  This is mainly
+intended as a workaround for https://github.com/magit/magit/issues/5176,
+in which case ?\N{ZERO WIDTH SPACE} is a good value.")
+
+(defun magit-insert-files (type fn &optional nogroup)
+  (when-let ((files (funcall fn)))
     (let* ((base (car magit-buffer-diff-files))
-           (base (and base (file-directory-p base) base)))
-      (magit-insert-section (assume-unchanged nil t)
-        (magit-insert-heading "Assume-unchanged files:")
-        (magit-insert-files files base)
+           (base (and base (file-directory-p base) base))
+           (title (symbol-name type)))
+      (magit-insert-section ((eval type) nil t)
+        (magit-insert-heading (length files)
+          (format "%c%s files"
+                  (capitalize (aref title 0))
+                  (substring title 1)))
+        (magit-insert-files-1 files base nogroup)
         (insert ?\n)))))
 
-(defun magit-insert-files (files directory)
-  (while (and files (string-prefix-p (or directory "") (car files)))
+(defun magit-insert-files-1 (files directory &optional nogroup)
+  (while (and files (or nogroup
+                        (not directory)
+                        (string-prefix-p directory (car files))))
     (let ((dir (file-name-directory (car files))))
-      (if (equal dir directory)
+      (if (or nogroup (equal dir directory))
           (let ((file (pop files)))
             (magit-insert-section (file file)
+              (when magit-file-section-indent
+                (insert magit-file-section-indent))
               (insert (propertize file 'font-lock-face 'magit-filename) ?\n)))
         (magit-insert-section (file dir t)
+          (when magit-file-section-indent
+            (insert magit-file-section-indent))
           (insert (propertize dir 'file 'magit-filename) ?\n)
           (magit-insert-heading)
-          (setq files (magit-insert-files files dir))))))
+          (setq files (magit-insert-files-1 files dir))))))
   files)
 
 ;;; _
